@@ -11,7 +11,7 @@ let%expect_test "exercise cutoff code path" =
       arr2
       ~f:(fun (i, j) ->
         assert (arr1.(i) = arr2.(j));
-        print_s [%sexp (arr1.(i) : int), (i, j : int * int)])
+        print_s [%sexp (arr1.(i) : int), ((i, j) : int * int)])
       ~hashable:(module Int)
   in
   do_diff ~cutoff:None;
